program Yac2Squeeze;

{$R *.res}

uses
  Windows, SysUtils, Classes, Registry, Sockets, HttpApp, Core;

const
  TITLE          = 'YAC to Squeeze';

  BANNER_CALL    = '********* THE PHONE IS RINGING *********';
  BANNER_MESSAGE = '*********** INCOMING MESSAGE ***********';
  CALL_SIGNATURE = '@CALL';

  MIN_DURATION   = 0;
  MAX_DURATION   = 300;

  MIN_BRIGHTNESS = 0;
  MAX_BRIGHTNESS = 4;

  MAX_PLAYERS    = 16;

  TCP_TIMEOUT    = 1000;

  DEBUG_LOG      = 'Yac2Squeeze-debug.txt';
  CRLF           = #13#10;

type
  TConfig = class
    Server,                // Server IP address or host name
    ServerPort,            // Port number of the media server CLI

    ServerUser,            // username (if security is enabled)
    ServerPass,            // password (if security is enabled)

    BannerCall,            // Banner text for incoming calls
    BannerMsg: string;     // Banner text for messages

    Display,               // Display duration in seconds
    Pause,                 // Pause duration in seconds (-1 = infinite, 0 = no pause, cannot be less than display duration)
    Brightness: integer;   // Brightness level when displaying call info (0 = no change)

    ShowError: boolean;    // Show error messages

    procedure Clear;
    procedure Load;

    constructor Create;
    destructor Destroy; override;
  end;

  TPlayer = record
    id: string;

    Power: boolean;        // Is player powered on?
    Brightness: integer;   // Player brightness level
  end;

var
  Config: TConfig;
  Players: array [1..MAX_PLAYERS] of TPlayer;
  cli: TTcpClient;
  params: string;
  count: integer;
  i: integer;
  line1, line2a, line2b: string;
  debug_fn: string;

function GetAllParams(var params: string): boolean;
var
  i: integer;
begin
  params := '';

  for i := 1 to ParamCount do
    params := params + ParamStr(i) + ' ';

  params := Trim(params);

  result := params <> '';
end;

procedure WriteToLog(s: string; timestamp: boolean = false; overwrite: boolean = false);
var
  tf: textfile;
begin
  {$IFDEF DEBUG}
  Assign(tf, debug_fn);

  if overwrite then
    Rewrite(tf)
  else
    Append(tf);

  if timestamp then
    s := FormatDateTime('hh:nn:ss.zzz  ', Now) + s;

  Writeln(tf, s);
  CloseFile(tf);
  {$ENDIF}
end;

// Issues a command via the CLI and returns only the
// answer portion of the response
function ServerCommand(command: string): string;
var
  response: string;
  qmark: integer;
begin
  WriteToLog(Format('Command:  %s', [HttpDecode(command)]), true);

  cli.Sendln(command);
  response := cli.Receiveln;

  // If we are asking server a question, parse the response
  qmark := Pos('?', command);
  if qmark > 0 then begin
    result := Copy(response, qmark, Length(response) - qmark + 1);

    if HttpDecode(result) = '?' then
      // Server didn't understand question
      raise Exception.Create('The media server responded unexpectedly. Try upgrading your media server software.');
   end
  else begin
    result := response;
  end;

  WriteToLog(Format('Response: %s', [HttpDecode(result)]), true);
  WriteToLog('');
end;

{ TConfig }

constructor TConfig.Create;
begin
  inherited;
end;

destructor TConfig.Destroy;
begin
  inherited;
end;

procedure TConfig.Clear;
begin
  Server     := 'localhost';
  ServerPort := '9090';
  ServerUser := '';
  ServerPass := '';

  BannerCall := BANNER_CALL;
  BannerMsg  := BANNER_MESSAGE;

  Display    := 30;
  Pause      := -1;
  Brightness := 0;

  ShowError  := true;
end;

procedure TConfig.Load;
var
  reg: TRegistry;

  function ReadRegString(v: string; d: string): string;
  begin
    try
      result := reg.ReadString(v);
    except
      result := d;
    end;

    if result = '' then
      result := d;
  end;

begin
  reg         := TRegistry.Create;
  reg.RootKey := HKEY_LOCAL_MACHINE;

  if reg.OpenKey('\SOFTWARE\Yac2Squeeze', false) then begin
    Server     := ReadRegString('Server',     Server);
    ServerPort := ReadRegString('ServerPort', ServerPort);
    ServerUser := ReadRegString('ServerUser', ServerUser);
    ServerPass := ReadRegString('ServerPass', ServerPass);

    BannerCall := ReadRegString('BannerCall', BannerCall);
    BannerMsg  := ReadRegString('BannerMsg',  BannerMsg);

    try Display    := StrToInt (ReadRegString('Display',    IntToStr(Display)));    except end;
    try Pause      := StrToInt (ReadRegString('Pause',      IntToStr(Pause)));      except end;
    try Brightness := StrToInt (ReadRegString('Brightness', IntToStr(Brightness))); except end;
    try ShowError  := StrToBool(ReadRegString('ShowError',  BoolToStr(ShowError))); except end;

    // Sanity checks
    if Display < MIN_DURATION then
      Display := MIN_DURATION;

    if Display > MAX_DURATION then
      Display := MAX_DURATION;

    if Pause > MAX_DURATION then
      Pause := MAX_DURATION;

    if Brightness < MIN_BRIGHTNESS then
      Brightness := MIN_BRIGHTNESS;

    if Brightness > MAX_BRIGHTNESS then
      Brightness := MAX_BRIGHTNESS;

    // Pause time cannot be less than display duration
    if (Pause > 0) and (Pause < Display) then
      Pause := Display;

    reg.CloseKey;
  end;

  reg.Free;
end;

begin
  // Allow only one instance
  CreateMutex(nil, true, TITLE);
  if GetLastError <> NO_ERROR then
    exit;

  debug_fn := ExtractFileDir(ParamStr(0));
  if (Length(debug_fn) > 0) and (Copy(debug_fn, Length(debug_fn), 1) <> '\') then
    debug_fn := debug_fn + '\';
  debug_fn := debug_fn + DEBUG_LOG;

  cli     := TTcpClient.Create(nil);
  Config  := TConfig.Create;
  Config.Clear;

  try
    WriteToLog(GetAppDescription + ' version ' + GetVersionString + ' (' + GetRawVersionString + '), compiled ' + GetCompileString, false, true);
    WriteToLog(GetCopyrightString);
    WriteToLog(GetWebAddress + '  ' + GetEmailAddress);
    WriteToLog('');
    WriteToLog(Format('Current Time: %s', [DateTimeToStr(Now)]));
    WriteToLog('');

    if GetAllParams(params) then begin
      WriteToLog(Format('Params: %s', [params]));
      WriteToLog('');

      Config.Load;

      WriteToLog(Format('Config.Server:      %s', [Config.Server]));
      WriteToLog(Format('Config.ServerPort:  %s', [Config.ServerPort]));
      WriteToLog(Format('Config.ServerUser:  %s', [Config.ServerUser]));
      WriteToLog(Format('Config.ServerPass:  %s', [Config.ServerPass]));
      WriteToLog(Format('Config.BannerCall:  %s', [Config.BannerCall]));
      WriteToLog(Format('Config.BannerMsg:   %s', [Config.BannerMsg]));
      WriteToLog(Format('Config.Display:     %d', [Config.Display]));
      WriteToLog(Format('Config.Pause:       %d', [Config.Pause]));
      WriteToLog(Format('Config.Brightness:  %d', [Config.Brightness]));
      WriteToLog(Format('Config.ShowError:   %s', [BoolToStr(Config.ShowError, true)]));
      WriteToLog('');

      cli.RemoteHost := Config.Server;
      cli.RemotePort := Config.ServerPort;

      cli.Connect;

      if cli.Connected then begin
        // Authenticate
        ServerCommand(Format('login %s %s', [Config.ServerUser, Config.ServerPass]));

        // If authentication fails, the server disconnects. The next command will fail
        try
          ServerCommand('version ?');
        except
          raise Exception.Create('Media server authentication failure. Check the ServerUser and ServerPass Registry values.');
        end;

        // Enumerate all players
        count := StrToInt(ServerCommand('player count ?'));
        for i := 1 to count do begin
          if i > MAX_PLAYERS then begin
            WriteToLog(Format('Warning: maximum players (%d) exceeded.', [MAX_PLAYERS]), true);
            WriteToLog('');

            count := MAX_PLAYERS;
            break;
          end;

          Players[i].id := ServerCommand(Format('player id %d ?', [i - 1]));
        end;

        // Get current state of players
        for i := 1 to count do begin
          Players[i].power  := ServerCommand(Format('%s power ?', [Players[i].id])) = '1';
          if Config.Brightness > 0 then
            Players[i].Brightness := StrToInt(ServerCommand(Format('%s playerpref power' + IIf(Players[i].power, 'On', 'Off') + 'Brightness ?', [Players[i].id])));
        end;

        // Pause, only if device is on and playing
        if Config.pause <> 0 then
          for i := 1 to count do
            if Players[i].power then
              if ServerCommand(Format('%s mode ?', [Players[i].id])) = 'play' then
                ServerCommand(Format('%s pause 1', [Players[i].id]));

        // Adjust brightness on player
        if Config.Brightness > 0 then
          for i := 1 to count do
            if Players[i].Brightness <> Config.Brightness then
              ServerCommand(Format('%s playerpref power' + IIf(Players[i].power, 'On', 'Off') + 'Brightness %d', [Players[i].id, Config.Brightness]));

        // Display message
        if Config.Display > 0 then begin
          if (Length(params) > Length(CALL_SIGNATURE)) and (Copy(params, 1, Length(CALL_SIGNATURE)) = CALL_SIGNATURE) and (Pos('~', params) > 0) then begin
            Delete(params, 1, Length(CALL_SIGNATURE));

            i := Pos('~', params);

            line1  := HttpEncode(Config.BannerCall);
            line2a := HttpEncode(Copy(params, 1, i - 1));
            line2b := HttpEncode('  ' + Copy(params, i + 1, Length(params) - i));
           end
          else begin
            line1  := HttpEncode(Config.BannerMsg);
            line2a := HttpEncode(params);
            line2b := '';
          end;

          // HttpEncode replaces spaces with plus signs, change those to %20's
          line1  := StringReplace(line1,  '+', '%20', [rfReplaceAll]);
          line2a := StringReplace(line2a, '+', '%20', [rfReplaceAll]);
          line2b := StringReplace(line2b, '+', '%20', [rfReplaceAll]);

          for i := 1 to count do
            ServerCommand(Format('%s display %s %s %d', [Players[i].id, line1, line2a + line2b, Config.Display]));
        end;

        // Sleep to revert brightness and/or unpause players
        // If neither is the case, we don't need to sleep and can just exit the program
        if (Config.Brightness > 0) or (Config.pause > 0) then begin
          WriteToLog(Format('Sleeping for %d seconds (brightness and/or pause state may have been changed).', [Config.Display]), true);
          WriteToLog('');

          Sleep(Config.Display * 1000);

          // Reset brightness on player
          if Config.Brightness > 0 then
            for i := 1 to count do
              if Players[i].Brightness <> Config.Brightness then
                ServerCommand(Format('%s playerpref power' + IIf(Players[i].power, 'On', 'Off') + 'Brightness %d', [Players[i].id, Players[i].Brightness]));

          // Now wait a bit more to unpause players (this is why pause time must be >= display time)
          if Config.pause > 0 then begin
            if (Config.pause - Config.Display) > 0 then begin
              WriteToLog(Format('Sleeping for %d more seconds (pause time is set longer than display time).', [Config.pause - Config.Display]), true);
              WriteToLog('');

              Sleep((Config.pause - Config.Display) * 1000);
            end;

            // Unpause player if (still) powered on and paused
            for i := 1 to count do
              if ServerCommand(Format('%s power ?', [Players[i].id])) = '1' then // do not use Players[i].power here, we want to check actual current status
                if ServerCommand(Format('%s mode ?', [Players[i].id])) = 'pause' then
                  ServerCommand(Format('%s pause 0', [Players[i].id]));
          end;
        end;

        cli.Disconnect;
      end;

      Config.Free;
      cli.Free;
    end;

    WriteToLog('Finished.', true);
  except
    on E: Exception do begin
      if Config.ShowError then
        MessageBox(0, PChar(E.Message), PChar(TITLE), MB_ICONWARNING or MB_OK);

      WriteToLog(Format('Exception: %s', [E.Message]), true);
      WriteToLog('');
    end;
  end;
end.
